/* ----------------------------------------------------------
%   (C)1993,1994,1995 Institute for New Generation Computer Technology
%       (Read COPYRIGHT for detailed information.)
%   (C)1996, 1997, 1998, 1999 Japan Information Processing Development Center
%       (Read COPYRIGHT-JIPDEC for detailed information.)
----------------------------------------------------------- */

#include <klic/basic.h>  /* fatal, fatalf, klic_fprintf */
#include <klic/struct.h>
#include <klic/primitives.h>  /* malloc_check, realloc_check */
#include "timing.h"
#include <stdio.h>
#include <klic/functorstuffs.h>  /* arityof */
#include "gobj.h"
#include <klic/susp.h>
#include "trace.h"  /* trace_deadlock */

#ifdef DIST
#include "interpe.h"
#endif

#ifdef SHM
#include "shm.h"
#endif  /* SHM */

#define generic_gc(obj)  (method_table_of(obj)->gc((obj), heapp()))

extern const struct predicate queue_empty_pred;

extern struct goalrec goal_queue_tail;

static q** gcstack;
static q** gcsp0;
static q** gcmax0;
static unsigned long gcstack_size;
static unsigned long copied_susp;
static struct suspended_goal_rec* suspended_goal_list0;

extern q** gcsp(void){ return gcsp0; }
extern q** gcmax(void){ return gcmax0; }
extern struct suspended_goal_rec* suspended_goal_list(void){ return suspended_goal_list0; }
extern void set_gcsp(q** pp){ gcsp0 = pp; }
extern void set_suspended_goal_list(struct suspended_goal_rec* p){ suspended_goal_list0 = p; }

static Inline void
flip_spaces()
{
  declare_globals;
  q* tempp = new_space_top();
  unsigned long temps = new_space_size();
  set_new_space_top(old_space_top());
  set_new_space_size(old_space_size());
  set_old_space_top(tempp);
  set_old_space_size(temps);
}

static void
make_larger_stack(void)
{
  declare_globals;
  q** newstack;
  gcstack_size *= 2;
  newstack = (q**) realloc_check((void*) gcstack, gcstack_size*sizeof(q*));
  set_gcsp(newstack + (gcsp() - gcstack));
  gcstack = newstack;
  gcmax0 = newstack + gcstack_size;
}

#ifdef SHM
extern q**
make_shm_larger_stack(sp)
  q** sp;
{
  declare_globals;
  q** newstack;
  shm_gcstack_size *= 2;
  newstack = (q**) realloc_check((void*) shm_gcstack, shm_gcstack_size*sizeof(q*));
  sp = newstack + (sp - shm_gcstack);
  shm_gcstack = newstack;
  shm_gcmax = newstack + shm_gcstack_size;
  return sp;
}

static void push_shm_stack(q* addr)
{
  if(shm_sp == shm_gcmax)
    shm_sp = make_shm_larger_stack(shm_sp);
  *shm_sp = addr;
  shm_sp++;
}
#endif /* SHM */

static int within_new_space(q* x)
{
  return (unsigned long)x - (unsigned long)new_space_top() < new_space_size();
}

static int within_old_space(q* x)
{
  return (unsigned long)x - (unsigned long)old_space_top() < old_space_size();
}

extern void push_gc_stack(q* addr)
{
  if(gcsp() == gcmax())
    make_larger_stack();
  *gcsp() = addr;
  set_gcsp(gcsp() + 1);
}

#ifdef SHM
#define reserve_copy(from, to) \
if( from == makeref(&from) ){ \
  to = from = makeref(&to); \
}else{ \
  to = from; \
  if( !isatomic(from) ){ \
    if( is_shma(from) ){ \
      push_shm_stack(&to); \
    }else if( within_old_space(from) ){ \
      from = makeref(&to); \
      push_gc_stack(&to); \
    } \
  } \
}
#else /* not SHM */
#define reserve_copy(from, to) \
if( from == makeref(&from) ){ \
  to = from = makeref(&to); \
}else{ \
  to = from; \
  if( !isatomic(from) && within_old_space(from) ){ \
    from = makeref(&to); \
    push_gc_stack(&to); \
  } \
}
#endif /* not SHM */

#define copy_one_goal(goal, susp)		\
do{							\
  struct goalrec* og = (goal);				\
  int n = og->pred->arity;				\
  struct goalrec* ng = (struct goalrec*) klic_alloc(n + 2); \
  ng->next = og->next;					\
  og->next = ng;					\
  ng->pred = og->pred;					\
  og->pred = 0;						\
  while( --n >= 0 ){					\
    reserve_copy(og->args[n], ng->args[n]); \
  }							\
  (goal) = ng;						\
  if( susp ) copied_susp++;				\
}while(0)

static void
copy_terms(void)
{
  declare_globals;
  while( gcsp() > gcstack ){
    q* addr;
    q obj;

    set_gcsp(gcsp() - 1);
    addr = *gcsp();
    obj = *addr;

  loop:
#ifdef SHM
    if( is_shma(obj) ){
      *addr = obj;
      if( ptagof(obj) != ATOMIC ){
	push_shm_stack(addr);
      }
      continue; /* while stack rests */
    }
#endif

    switch( ptagof(obj) ){
    case ATOMIC:
      *addr = obj;
      break;

    case CONS:
    cons_case:
      if( !within_old_space((q*) obj) ){
	*addr = obj;
      }else{
	q cdr = cdr_of(obj);
	if( isstruct(cdr) && within_new_space((q*) cdr) ){
	  *addr = cdr;
	}else{
	  q newcons = makecons(klic_alloc(2));
	  reserve_copy(car_of(obj), car_of(newcons));
	  *addr = cdr_of(obj) = newcons;
	  if( isatomic(cdr) ){
	    cdr_of(newcons) = cdr;
	  }else if( cdr == makeref(&cdr_of(obj)) ){
	    cdr_of(newcons) = makeref(&cdr_of(newcons));
	  }else{
	    addr = &cdr_of(newcons);
	    obj = cdr;
	    goto loop;
	  }
	}
      }
      break;

    case FUNCTOR:
    functor_case:
      if( !within_old_space((q*) obj) ){
	*addr = obj;
      }else{
	q f = functor_of(obj);
	if( isstruct(f) ){
	  *addr = f;
	}else if( isref(f) ){
	  struct data_object* oldobj = (struct data_object*) functorp(obj);
	  q* newobj;
	  newobj = generic_gc(oldobj);
	  *addr = functor_of(obj) = makefunctor(newobj);
	}else{
	  int k = arityof(f);
	  q newfunct = makefunctor(klic_alloc(k + 1));
	  *addr = functor_of(obj) = newfunct;
	  functor_of(newfunct) = f;
	  do{
	    k--;
	    reserve_copy(arg(obj,k), arg(newfunct,k));
	  }while( k > 0 );
	}
      }
      break;

    case VARREF:
      {
	q value;
      deref:
	value = derefone(obj);
	switch( ptagof(value) ){
	case ATOMIC:
	  *addr = value;
	  break;
	case FUNCTOR:
	  obj = value;
	  goto functor_case;
	case CONS:
	  if( !within_new_space((q*) value) ){
	    obj = value;
	    goto cons_case;
	  }
	  *addr = makeref(&cdr_of(value));
	  break;
	case VARREF:
	  if( derefone(value) != obj ){
	    if( within_old_space((q*) value) ){
	      obj = value;
	      goto deref;
	    }
	    *addr = value;
#ifdef SHM
	    if( is_shma(value) ) push_shm_stack(addr);
#endif
	  }else if( value == obj ){
	    if( within_new_space(addr) ){
	      *addr = derefone(obj) = makeref(addr);
	    }else{
	      q* hp = klic_alloc(1);
	      *addr = derefone(obj) = *hp = makeref(hp);
	    }
	  }else{
	    struct susprec* s = suspp(value);
	    if( is_generator_susp(s->u) ){
	      struct generator_object* oldobj;
	      struct generator_susp* newgsusp;
	      struct generator_susp* gsusp = generator_suspp(s);
	      q newplace;
	      q newvar = *addr = derefone(obj) = makeref(klic_alloc(1));
	      q* hp = klic_alloc(sizeof(struct generator_susp) / sizeof(q));
	      newgsusp = generator_suspp(hp);
	      derefone(newvar) = makeref(newgsusp);
	      newgsusp->backpt = makeref(newvar);
	      oldobj = untag_generator_susp(gsusp->u.o);
	      newplace = (q) (oldobj->method_table);
	      if( isstruct(newplace) ){
		/* already copied */
		newgsusp->u.o = tag_generator_susp(functorp(newplace));
	      }else{
		/* not yet copied */
		struct generator_object* newobj;
		newobj = (struct generator_object*) generic_gc(oldobj);
		oldobj->method_table =
		  (struct generator_object_method_table*) makefunctor(newobj);
		newgsusp->u.o = tag_generator_susp(newobj);
	      }
	    }else{
	      struct hook* second_hook = s->u.first_hook.next;
	      struct hook* h = second_hook;
	      struct hook dummy;
	      struct hook* last = &dummy;
	      union goal_or_consumer lastu;
	      q newvar;

	      /* make a new variable, anyway */
	      q* hp = klic_alloc(1);
	      newvar = *addr = derefone(obj) = *hp = makeref(hp);
	      lastu.l = 0;
	      do{
		union goal_or_consumer u = h->u;
		if( u.l != 0 ){
		  union goal_or_consumer nu;
		  nu.l = 0;
		  if( !is_consumer_hook(u) ){
		    /* suspended goal */
		    if( u.g->pred == 0 ){
		      nu.g = u.g->next;
		    }else{
		      if(isref(u.g->next)) goto invalid_hook;
		      nu.g = u.g;
		      copy_one_goal(nu.g, 1);
		    }
		  }else{
		    /* consumer object */
		    q newplace = (q) (untag_consumer_hook(u.o)->method_table);
		    if( isstruct(newplace) ){
		      nu.o = tag_consumer_hook(functorp(newplace));
		    }else{
		      struct consumer_object* newobj;
		      newobj = (struct consumer_object*)
			  generic_gc(untag_consumer_hook(u.o));
		      untag_consumer_hook(u.o)->method_table =
			(struct consumer_object_method_table*)
			  makefunctor(newobj);
		      nu.o = tag_consumer_hook(newobj);
		    }
		  }
		  if( lastu.l != 0 ){
		    q* hp = klic_alloc(sizeof(struct hook) / sizeof(q));
		    struct hook* nh = (struct hook*) hp;
		    nh->u = lastu;
		    last->next = nh;
		    last = nh;
		  }
		  lastu = nu;
		}
	      invalid_hook:
		h = h->next;
	      }while( h != second_hook );
	      if( lastu.l != 0 ){
		q* hp = klic_alloc(sizeof(struct susprec) / sizeof(q));
		struct susprec* ns = (struct susprec*) hp;
		last->next = &ns->u.first_hook;
		ns->backpt = newvar;
		ns->u.first_hook.next = dummy.next;
		ns->u.first_hook.u = lastu;
		derefone(newvar) = (q) ns;
	      }
	    }
	  }
	  break;
	}  /* switch ptagof(value) */
      }
      break;
    }  /* switch ptagof(obj) */
  }  /* while stack rests */
}

static struct goalrec*
copy_one_queue(qp)
  struct goalrec* qp;
{
  declare_globals;
  struct goalrec* last;
  struct goalrec* next;

  /* Copy queue in reverse order */
  /* By this, variables will have better chance to be allocated */
  /* within the goal records or structures that'll be read after */
  /* their instantiation */

  /* First, reverse the goal queue */
  for( last=NULL; qp != &goal_queue_tail; last=qp, qp=next ){
    next = qp->next;
    qp->next = last;
  }
  /* Then copy and rearrange the goal queue */
  qp = last;
  last = &goal_queue_tail;
  for( ; qp != NULL; qp=next ){
    next = qp->next;
#ifdef SHM
    if( is_shma(qp) ){
      qp->next = last;
      last = qp;
      continue; /* for qp */
    }
#endif
    copy_one_goal(qp, 0);
    copy_terms();
    qp->next = last;
    last = qp;
  }
  return last;
}

static struct goalrec*
collect_garbage(qp)
  struct goalrec* qp;
{
  declare_globals;
  int k;
  unsigned long nsize, osize;
  struct prioqrec* pq = prioq().next;

#ifdef DIST
  if( node_wtc == 0 ){
    fatal("invalid WTC in gc");
  }
#endif

  if( gctimes==0 ){
    /* allocate GC stack on first GC */
    gcstack_size = GCSTACKSIZE;
    gcstack = (q**) malloc_check(gcstack_size*sizeof(q*));
    set_gcsp(gcstack);
    gcmax0 = gcstack + gcstack_size;
#ifdef SHM
    shm_gcstack_size = GCSTACKSIZE;
    shm_gcstack = (q**) malloc(shm_gcstack_size*sizeof(q*));
    shm_sp = shm_gcstack;
    shm_gcmax = shm_gcstack+shm_gcstack_size;
  }else{
    shm_sp = shm_gcstack;
#endif
  }

#ifdef SHM
  /* copy into Shared-memory Generator hook data */
  {
    TADDRtbl* nptr;
    TADDRtbl* sptr = &ADDRtbl;
    for( sptr=sptr->next; sptr != &ADDRtbl; sptr=nptr ){
      nptr = sptr->next;
      switch( ptagof(sptr->localA) ){
      case CONS:
      case FUNCTOR: /* generator hook */
	{
	  struct generator_object* addi;
	  Shvar* objp;
	  q temp;
	Re_try:
	  temp = derefone(sptr->globalA);
	  if( !isref(temp) ){ break; /* switch */ }
	  if( derefone(temp) != (q) sptr->globalA ){
	    sptr->globalA = (q*) temp;
	    goto Re_try;
	  }
	  addi = n_lock((q) sptr->globalA, temp);
	  if( derefone(sptr->globalA) != temp ){ goto Re_try; }
	  objp = (Shvar*) untag_generator_susp(addi);
	  if( is_genhook(objp->chain) ){
	    q tempg;
	    shm_arg_copy(&sptr->localA, &tempg);
	    klic_barrier();
	    *(sptr->globalA) = tempg;
	    free_local_tbl(sptr);
	  }else{
	    n_unlock(temp, addi);
	  }
	}
      }
    }
  }
#endif /* SHM */

  set_heaptop(new_space_top());
  set_heapp(new_space_top());
  set_real_heapbytesize(new_space_size());
  set_real_heaplimit(heapp() + heapsize());
  heaplimit = (interrupt_off ? real_heaplimit() : 0);
  set_heapbottom(real_heaplimit() + incrementsize());

  for( k=0; k<num_gc_hooks(); k++ ){
    ((gc_t) gc_hook_table(k))();
  }

  for( ; pq->prio >= 0; pq = pq->next ){
    pq->q = copy_one_queue(pq->q);
  }
  qp = copy_one_queue(qp);

#ifdef SHM
  {
    TADDRtbl* nptr;
    TADDRtbl* sptr = &ADDRtbl;
    for( sptr=sptr->next; sptr!=&ADDRtbl; sptr=nptr ){
      nptr = sptr->next;
      switch( ptagof(sptr->localA) ){
      case CONS:
      case FUNCTOR:  /* generator hook but anybody reqested */
        push_gc_stack((q*) &sptr->localA);
	copy_terms();
        push_shm_stack(&sptr->globalA);
        break;
      case ATOMIC: /* genarator object (distributed interface) */
	{
	  q wk = (q) untag_local(sptr->localA);
	  if( !derefone(wk) ){ /* consumer */
	    /* a following line is for debugging of reverse pointer problem. */
	    static int hirata_bug1;

	    q top = (q) &(sptr->localA);
	    derefone(wk) = top;
	    derefone(top) = wk;
	    push_gc_stack(&top);

	    /* patch for debugging of hirata problem */
	    hirata_bug1 = 1;
	    copy_terms();
	    hirata_bug1 = 0;

	    wk = derefone(top);
	    if( wk == top )
	      goto REM_HOOK;
	    sptr->localA = (q*) tag_local(wk);
	    derefone(wk) = 0;
	  }else{ /* generator */
	    push_gc_stack(&wk);
	    copy_terms();
	    sptr->localA = (q*) tag_local(wk);
	  }
	  push_shm_stack(&sptr->globalA);
	  break;
	}
      default: /* normal goal */
	{
	  struct goalrec* wqp = (struct goalrec*) sptr->localA;
	  if( !wqp ){ /* skip */
	  }else if( !wqp->pred ){
	    sptr->localA = (q*) wqp->next;
	  }else if( isint(wqp->next) ){
	    copy_one_goal(wqp, 1);
	    sptr->localA = (q*) wqp;
	    copy_terms();
	    if( isref(sptr->globalA) ){
	      for(;;){
		q ww = derefone(sptr->globalA);
		if( !isref(ww) || (q)sptr->globalA == derefone(ww) ) break;
		sptr->globalA = (q*) ww;
	      }
	    }
	    push_shm_stack(&sptr->globalA);
	  }else{
	  REM_HOOK:
	    sptr->localA = 0;
	    /* removes a hook record */
	    {
	      struct generator_object* addi;
	      q sv;
	      Shvar* objp;
	      Sinfo *hk, *bsi, *si;
	      q var = (q) sptr->globalA;
	      sv = derefone(var);
	      if( !isref(sv) || derefone(sv) != var ){ goto REM_skip; }
	      addi = n_lock(var, sv);
	      if( derefone(var) != sv ) goto REM_skip;
	      objp = (Shvar*) untag_generator_susp(addi);
	      hk = objp->chain;
	      if( hk->indp == sptr ){
		objp->chain = hk->next;
		free_local_tbl(sptr);
	      }else{
		for( bsi=hk, si=bsi->next; si != NULL; bsi=si, si=si->next ){
		  if( si->indp == sptr ){
		    bsi->next = si->next;
		    free_local_tbl(sptr);
		    break; /* for si */
		  }
		}
	      }
	      n_unlock(sv, addi);
	    REM_skip:
	      ;
	    }
	  }
	}
      }
    }
  }
#endif /* SHM */

  {
    struct suspended_goal_rec* sgl;
    struct suspended_goal_rec** sgl_tail = &suspended_goal_list0;
    struct goalrec* dead_goal = NULL;
    if( copied_susp != suspensions() - resumes() ){
      klic_fprintf(stderr,
		   "%d perpetually suspending goals found\n",
		   suspensions() - resumes() - copied_susp);
    }

    /* First, we will copy the surface of suspended goal list.
     * This is needed to distinguish resumed goals and
     * goals copied while copying other goals in this list */
    for( sgl = suspended_goal_list(); sgl != NULL; sgl = sgl->next ){
      if( sgl->goal->pred == 0 || !isref(sgl->goal->next) ){
	/* already copied or not resumed yet */
	q* hp = klic_alloc(sizeof(struct suspended_goal_rec) / sizeof(q));
	struct suspended_goal_rec* newsgr = (struct suspended_goal_rec*) hp;
	*sgl_tail = newsgr;
	newsgr->goal = sgl->goal;
	sgl_tail = &newsgr->next;
      }
    }
    *sgl_tail = NULL;

    /* Next, we will copy the suspended goals */
    for( sgl = suspended_goal_list(); sgl != NULL; sgl = sgl->next ) {
      if( sgl->goal->pred == 0 ){
	sgl->goal = sgl->goal->next;
      }else{
	/* not copied yet */
	copy_one_goal(sgl->goal, 0);
	copy_terms();
	dead_goal = sgl->goal;
      }
    }
    if( dead_goal != 0 ){
      /* we have to make the dead goal look like a normal ready queue */
      q* hp = klic_alloc(sizeof(struct goalrec) / sizeof(q));
      ((struct goalrec*) hp)->pred = &queue_empty_pred;
      dead_goal->next = (struct goalrec*) hp;
      trace_deadlock(dead_goal);
    }
  }

  if( copied_susp != suspensions() - resumes() ){
    fatal("Perpetually suspending goal(s) found during GC");
  }

  for( k=0; k<num_after_gc_hooks(); k++ ){
    ((gc_t) after_gc_hook_table(k))();
  }

  return qp;
}

/*
  interface routine for copying one term
*/

extern void
copy_one_term(term)
  q* term;
{
  declare_globals;
  push_gc_stack(term);
  copy_terms();
}

/*
  for generic object
*/
extern q
general_gc(term)
  q* term;
{
  declare_globals;

  push_gc_stack(term);
  copy_terms();
  return *term;
}


extern struct goalrec*
klic_gc(qp)
  struct goalrec* qp;
{
  declare_globals;
  timerstruct before, after;
  static int lastgc_dangerous = 0;
  static int make_heap_larger = 0; /* make heap larger in the next GC */
  q* new_new_space_top;
  unsigned long bytesize;
  int k;
  if( measure_gc ) measure(before);

 gc_again:
  if( make_heap_larger ){
    q* new_old_space_top;
    double_heapsize();
    bytesize = (heapsize() + incrementsize()) * sizeof(q);
    new_old_space_top = (q*) malloc(bytesize);
    if( new_old_space_top != NULL ){
      new_new_space_top = (q*) malloc(bytesize);
      if( new_new_space_top != NULL ){
	free(old_space_top);
	set_old_space_top(new_old_space_top);
	set_old_space_size(bytesize);
      }else{
	free(new_old_space_top);
	if( lastgc_dangerous ){
	  fatal("Not enough space collected and can't make heap larger");
	}
      }
    }
  }else if( lastgc_dangerous ){
    fatalf("Maximum heap size specified (%u words) has been used up",
	   maxheapsize());
  }
  flip_spaces();
  copied_susp = 0;		/* for perpetual suspension detection */
  qp = collect_garbage(qp);
  if( make_heap_larger ){
    free(old_space_top);
    set_old_space_top(new_new_space_top);
    set_old_space_size(bytesize);
  }
  make_heap_larger =
    (heapp()-new_space_top()+this_more_space() > heapsize()*maxactiveratio() &&
     heapsize() < maxheapsize());
  lastgc_dangerous = (real_heaplimit() < heapp()+this_more_space());
  if( lastgc_dangerous ){
    goto gc_again;
  }
  reset_this_more_space();
  gctimes++;
#ifdef SHM
  if( F_shm_gc ) qp = shm_gc(qp);
#endif
  if( measure_gc ){
    measure(after);
#ifdef GETRUSAGE
    gcums += diff_usec(ru_utime)/1000;
    gcsms += diff_usec(ru_stime)/1000;
#else
    gcums += (int) tick2msec(field_diff(tms_utime));
    gcsms += (int) tick2msec(field_diff(tms_stime));
#endif
  }
  return qp;
}
